"use strict";
exports.__esModule = true;
var ClientAPI_1 = require("../schemas/ClientAPI");
var models_1 = require("../common/models");
var util_1 = require("./util");
var debounce_1 = require("./debounce");
var DecisionTree_1 = require("./DecisionTree");
var SESSION_EXPIRES_AFTER = 24 * (60 * 60 * 1000);
var ExperimentOptions = (function () {
    function ExperimentOptions(bestOption, epsilon) {
        if (bestOption === void 0) { bestOption = undefined; }
        if (epsilon === void 0) { epsilon = 1; }
        this.bestOption = bestOption;
        this.epsilon = epsilon;
    }
    return ExperimentOptions;
}());
exports.ExperimentOptions = ExperimentOptions;
var Client = (function () {
    function Client(environment, appKey, then, outcomes) {
        var _this = this;
        this.environment = environment;
        this.appKey = appKey;
        this.experimentOptions = {};
        this.experiments = {};
        this.defaultCompletions = {};
        this.queuedCompletedExperiments = {};
        this.queuedStartedExperiments = {};
        this.preinitialized = false;
        this.serializeStateDebounced = debounce_1.debounce(function () {
            _this.log("Writing state", _this.serialized);
            try {
                var stateString = models_1.Convert.serializedStateToJson(_this.serialized);
                _this.environment.setLocalStorage(_this.storageKey("state"), stateString);
            }
            catch (e) {
                _this.error("Could not save state:", e.message);
            }
        }, 100);
        this.startExperimentsDebounced = debounce_1.debounce(function () {
            var experiments = util_1.mapObject(_this.queuedStartedExperiments, function (e) { return ({
                instanceKey: e.key,
                options: e.optionNames,
                pick: e.pick,
                pickedBest: e.pickedBest
            }); });
            _this.log("Starting experiments", experiments);
            _this.queuedStartedExperiments = {};
            var data = {
                version: 2,
                appKey: _this.appKey,
                experiments: experiments,
                ctx: _this.getClientContext()
            };
            _this.environment.http("POST", ClientAPI_1.startExperimentsURL, data, function () {
                return;
            }, function (e) { return _this.error("Failed to start experiments", e); });
        }, 100);
        this.completeExperimentsDebounced = debounce_1.debounce(function (then) {
            var experiments = util_1.getOwnPropertyValues(_this.queuedCompletedExperiments);
            _this.queuedCompletedExperiments = {};
            var experimentsByKey = {};
            for (var _i = 0, experiments_1 = experiments; _i < experiments_1.length; _i++) {
                var e = experiments_1[_i];
                if (e.payoff !== undefined) {
                    experimentsByKey[e.key] = { pick: e.pick, payoff: e.payoff };
                }
            }
            _this.log("Completing experiments", experimentsByKey);
            function callThen() {
                if (then !== undefined) {
                    then();
                }
            }
            _this.environment.http("POST", ClientAPI_1.completeExperimentsURL, {
                version: 1,
                appKey: _this.appKey,
                experiments: experimentsByKey
            }, function () { return callThen(); }, function (e) {
                _this.error("Failed to complete experiments", e);
                callThen();
            });
        }, 10);
        this.log("Initialize", appKey);
        this.preinitialized = outcomes !== undefined;
        this.serialized = {
            lastInitialized: 0,
            experimentPicks: {}
        };
        try {
            var stateString = environment.getLocalStorage(this.storageKey("state"));
            if (stateString !== undefined) {
                this.serialized = models_1.Convert.toSerializedState(stateString);
            }
        }
        catch (_a) { }
        var now = new Date().getTime();
        var beginNewSession = now - this.serialized.lastInitialized > SESSION_EXPIRES_AFTER;
        if (beginNewSession) {
            this.log("Starting a new session");
            this.serialized.experimentPicks = {};
        }
        this.serialized.lastInitialized = now;
        this.serializeStateDebounced();
        if (outcomes !== undefined) {
            this.finishInit(outcomes);
            return;
        }
        environment.http("GET", ClientAPI_1.outcomesUrl(appKey), undefined, function (o) {
            _this.log("Got outcomes", o);
            _this.finishInit(o);
            then(_this);
        }, function (e) {
            _this.error("Could not get outcomes", e);
            _this.finishInit({});
            then(_this);
        });
    }
    Client.prototype.log = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        (_a = this.environment).log.apply(_a, args);
        var _a;
    };
    Client.prototype.error = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        (_a = this.environment).error.apply(_a, args);
        var _a;
    };
    Client.prototype.getClientContext = function () {
        if (this.clientContext === undefined) {
            this.clientContext = { tzo: this.environment.getTimeZoneOffset() };
            var lang = this.environment.getLocalLanguage();
            if (lang !== undefined) {
                this.clientContext.lang = lang;
            }
        }
        return this.clientContext;
    };
    Client.prototype.startExperiment = function (theExperiment) {
        this.queuedStartedExperiments[theExperiment.name] = theExperiment;
        this.startExperimentsDebounced();
    };
    Client.prototype.completeExperiment = function (theExperiment, then) {
        this.queuedCompletedExperiments[theExperiment.name] = theExperiment;
        this.completeExperimentsDebounced(then);
    };
    Client.prototype.completeDefaults = function (score, then) {
        var completions = this.defaultCompletions;
        Object.getOwnPropertyNames(completions).forEach(function (name) { return completions[name].complete(score, then); });
    };
    Client.prototype.finishInit = function (outcomes) {
        var _this = this;
        try {
            var ctx_1 = this.getClientContext();
            Object.getOwnPropertyNames(outcomes).forEach(function (name) {
                if (_this.experimentOptions[name] !== undefined)
                    return;
                var _a = DecisionTree_1.lookupBestOption(ctx_1, outcomes[name]), option = _a.option, epsilon = _a.epsilon;
                _this.experimentOptions[name] = new ExperimentOptions(option, epsilon);
            });
            if (this.preinitialized) {
                this.environment.startHTMLExperiments();
                this.environment.autocomplete(function (payoff) { return _this.completeDefaults(payoff, undefined); });
            }
        }
        catch (e) {
            this.error("Couldn't finish init", e);
        }
    };
    Client.prototype.experiment = function (name, optionNames) {
        var ex = this.experiments[name];
        if (ex !== undefined)
            return ex;
        var options = this.experimentOptions[name];
        if (options === undefined) {
            options = this.experimentOptions[name] = new ExperimentOptions();
        }
        ex = this.experiments[name] = new Experiment(this, name, optionNames, options);
        this.defaultCompletions[name] = ex;
        this.startExperiment(ex);
        return ex;
    };
    Client.prototype.storageKey = function (path) {
        return "autotune.v1." + this.appKey + "." + path;
    };
    Client.prototype.loadPick = function (name) {
        return this.serialized.experimentPicks[name];
    };
    Client.prototype.savePick = function (name, pick) {
        this.serialized.experimentPicks[name] = pick;
        this.serializeStateDebounced();
    };
    return Client;
}());
exports.Client = Client;
var Experiment = (function () {
    function Experiment(client, name, optionNames, options) {
        this.client = client;
        this.name = name;
        this.optionNames = optionNames;
        this.options = options;
        this.key = util_1.uuidv4();
        var _a = this.options, bestOption = _a.bestOption, epsilon = _a.epsilon;
        var savedPick = this.client.loadPick(this.name);
        if (savedPick !== undefined && optionNames.indexOf(savedPick) !== -1) {
            this.pick = savedPick;
            this.pickedBest = savedPick === bestOption;
        }
        else {
            var pickRandom = bestOption === undefined ||
                optionNames.indexOf(bestOption) === -1 ||
                Math.random() < epsilon;
            var pick = void 0;
            if (pickRandom || bestOption === undefined) {
                pick = optionNames[Math.floor(Math.random() * optionNames.length)];
            }
            else {
                pick = bestOption;
            }
            this.client.savePick(this.name, pick);
            this.pick = pick;
            this.pickedBest = !pickRandom;
        }
    }
    Experiment.prototype.complete = function (payoff, then) {
        if (payoff === void 0) { payoff = 1; }
        this.payoff = payoff;
        this.client.completeExperiment(this, then);
    };
    return Experiment;
}());
exports.Experiment = Experiment;
//# sourceMappingURL=Client.js.map