"use strict";
exports.__esModule = true;
var Client_1 = require("./Client");
var ClientConfig_1 = require("../common/ClientConfig");
var models_1 = require("../common/models");
var validateUUID = require("uuid-validate");
var ClientAPI_1 = require("../schemas/ClientAPI");
jest.useFakeTimers();
var appKey = "abcde";
var experimentName = "ex";
var defaultClientContext = { tzo: -420, lang: "en" };
function failAndThrow(msg) {
    fail(msg);
    throw new Error(msg);
}
var TestEnvironment = (function () {
    function TestEnvironment(outcomes, clientContext, allowStartExperiments, allowSetLocalStorage, allowCompleteExperiments) {
        this.outcomes = outcomes;
        this.allowStartExperiments = allowStartExperiments;
        this.allowSetLocalStorage = allowSetLocalStorage;
        this.allowCompleteExperiments = allowCompleteExperiments;
        this.logs = [];
        this.errors = [];
        this.numOutcomesRequested = 0;
        this.htmlExperimentsStarted = false;
        this.autocompleteCalled = false;
        this.startExperimentsData = undefined;
        this.completeExperimentsData = undefined;
        this.localStorage = {};
        this.clientContext = Object.assign({}, clientContext);
    }
    TestEnvironment.testWithEnvironment = function (fullName, env, fn) {
        test(fullName, function () {
            var callback = jest.fn();
            var client = new Client_1.Client(env, appKey, callback);
            env.setClient(client);
            jest.advanceTimersByTime(150);
            expect(callback).toHaveBeenCalledTimes(1);
            expect(callback).toHaveBeenCalledWith(client);
            fn(env);
            jest.runAllTimers();
            expect(callback).toHaveBeenCalledTimes(1);
        });
    };
    TestEnvironment.test = function (name, settings, fn) {
        for (var _i = 0, _a = [false, true]; _i < _a.length; _i++) {
            var allowStartExperiments = _a[_i];
            for (var _b = 0, _c = [false, true]; _b < _c.length; _b++) {
                var allowSetLocalStorage = _c[_b];
                for (var _d = 0, _e = [false, true]; _d < _e.length; _d++) {
                    var allowCompleteExperiments = _e[_d];
                    var withs = [];
                    if (allowStartExperiments)
                        withs.push("startExperiments");
                    if (allowSetLocalStorage)
                        withs.push("local storage");
                    if (allowCompleteExperiments)
                        withs.push("completeExperiments");
                    var env = new TestEnvironment(settings.outcomes, settings.clientContext, allowStartExperiments, allowSetLocalStorage, allowCompleteExperiments);
                    var suffix = withs.length === 0 ? "" : " with " + withs.join(",");
                    var fullName = name + suffix;
                    TestEnvironment.testWithEnvironment(fullName, env, fn);
                }
            }
        }
    };
    TestEnvironment.prototype.setClient = function (client) {
        if (this.maybeClient !== undefined) {
            return failAndThrow("Client cannot be set twice");
        }
        this.maybeClient = client;
    };
    TestEnvironment.prototype.getClient = function () {
        if (this.maybeClient === undefined) {
            return failAndThrow("Client not set");
        }
        return this.maybeClient;
    };
    TestEnvironment.prototype.getOptions = function () {
        if (this.outcomes === undefined) {
            return failAndThrow("No options when no outcomes are defined");
        }
        var outcomes = this.outcomes[experimentName];
        if (outcomes === undefined) {
            return failAndThrow("Experiment not defined");
        }
        return outcomes.options;
    };
    TestEnvironment.prototype.log = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        this.logs.push(args);
    };
    TestEnvironment.prototype.error = function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        this.errors.push(args);
    };
    TestEnvironment.prototype.getTimeZoneOffset = function () {
        if (typeof this.clientContext.tzo !== "number") {
            throw new Error("tzo must be a number");
        }
        return this.clientContext.tzo;
    };
    TestEnvironment.prototype.getLocalLanguage = function () {
        if (this.clientContext.lang !== undefined && typeof this.clientContext.lang !== "string") {
            throw new Error("lang must be undefined or a string");
        }
        return this.clientContext.lang;
    };
    TestEnvironment.prototype.http = function (method, url, data, resolve, reject) {
        var _this = this;
        setTimeout(function () {
            if (method === "GET" && url.endsWith(appKey + ".json")) {
                _this.numOutcomesRequested += 1;
                if (_this.outcomes === undefined) {
                    reject(new Error("Simulated GET outcomes failure"));
                }
                else {
                    resolve(_this.outcomes);
                }
            }
            else if (method === "POST" && url === ClientAPI_1.startExperimentsURL) {
                if (data === undefined) {
                    return failAndThrow("No data given for startExperiments");
                }
                if (_this.startExperimentsData !== undefined) {
                    return failAndThrow("startExperiments posted more than once");
                }
                _this.startExperimentsData = data;
                if (_this.allowStartExperiments) {
                    resolve({});
                }
                else {
                    reject(new Error("Simulated POST startExperiments failure"));
                }
            }
            else if (method === "POST" && url === ClientAPI_1.completeExperimentsURL) {
                if (data === undefined) {
                    return failAndThrow("No data given for completeExperiments");
                }
                if (_this.completeExperimentsData !== undefined) {
                    return failAndThrow("completeExperiments posted more than once");
                }
                _this.completeExperimentsData = data;
                if (_this.allowCompleteExperiments) {
                    resolve({});
                }
                else {
                    reject(new Error("Simulated POST completeExperiments failure"));
                }
            }
            else {
                fail("Unexpected HTTP request: " + method + " to " + url);
            }
        }, 10);
    };
    TestEnvironment.prototype.getLocalStorage = function (key) {
        throw new Error("Method not implemented.");
    };
    TestEnvironment.prototype.setLocalStorage = function (key, value) {
        if (this.allowSetLocalStorage) {
            this.localStorage[key] = value;
        }
        else {
            throw new Error("Local storage simulated not to work");
        }
    };
    TestEnvironment.prototype.startHTMLExperiments = function () {
        this.htmlExperimentsStarted = true;
    };
    TestEnvironment.prototype.autocomplete = function (complete) {
        this.autocompleteCalled = true;
    };
    TestEnvironment.prototype.checkStartExperiments = function (options, pick, pickedBest) {
        var data = this.startExperimentsData;
        if (typeof data !== "object" || data === null) {
            return failAndThrow("Illegal startExperiments data");
        }
        expect(data.version).toBe(2);
        expect(data.appKey).toBe(appKey);
        expect(data.ctx).toEqual(this.clientContext);
        expect(Object.getOwnPropertyNames(data.experiments)).toEqual([experimentName]);
        var ex = data.experiments[experimentName];
        if (!validateUUID(ex.instanceKey, 4)) {
            fail("instanceKey is not a v4 UUID");
        }
        expect(ex.options).toEqual(options);
        expect(ex.pick).toBe(pick);
        if (pickedBest !== undefined) {
            expect(ex.pickedBest).toBe(pickedBest);
        }
        return ex.instanceKey;
    };
    return TestEnvironment;
}());
function makeOptions(n) {
    var options = [];
    for (var i = 0; i < n; i++) {
        options.push("o" + i.toString());
    }
    return options;
}
function makeOutcomes(n, tree) {
    var options = makeOptions(n);
    var outcomes = {};
    outcomes[experimentName] = { options: options, tree: tree };
    return outcomes;
}
function leaf(best) {
    return { best: best, eps: 0 };
}
function tzoLt(v, l, r) {
    return { at: "tzo", op: ClientConfig_1.Op.Lt, v: v, l: l, r: r };
}
function langEq(v, l, r) {
    return { at: "lang", op: ClientConfig_1.Op.Eq, v: v, l: l, r: r };
}
function checkInit(env) {
    expect(env.numOutcomesRequested).toBe(1);
}
function checkState(localStorage, pick) {
    var str = localStorage["autotune.v1." + appKey + ".state"];
    if (str === undefined) {
        return failAndThrow("Local storage state not set");
    }
    var state = models_1.Convert.toSerializedState(str);
    var now = new Date().getTime();
    expect(state.lastInitialized).toBeLessThanOrEqual(now);
    expect(state.lastInitialized).toBeGreaterThan(now - 1100);
    var keys = Object.getOwnPropertyNames(state.experimentPicks);
    if (pick === undefined) {
        expect(keys).toEqual([]);
    }
    else {
        expect(keys).toEqual([experimentName]);
        expect(state.experimentPicks[experimentName]).toBe(pick);
    }
}
function checkCompleteExperiments(data, instanceKey, pick, payoff) {
    if (typeof data !== "object" || data === null) {
        return failAndThrow("Illegal completeExperiments data");
    }
    expect(data.appKey).toBe(appKey);
    var experiments = {};
    experiments[instanceKey] = { pick: pick, payoff: payoff };
    expect(data.experiments).toEqual(experiments);
}
TestEnvironment.test("can init when network fails", { clientContext: defaultClientContext }, function (env) {
    checkInit(env);
    jest.advanceTimersByTime(200);
    expect(env.errors.length).toBeGreaterThan(0);
    expect(env.startExperimentsData).toBe(undefined);
});
TestEnvironment.test("can init with network", { outcomes: {}, clientContext: defaultClientContext }, function (env) {
    checkInit(env);
    jest.advanceTimersByTime(200);
    if (env.allowSetLocalStorage) {
        expect(env.errors.length).toBe(0);
        checkState(env.localStorage, undefined);
    }
    expect(env.startExperimentsData).toBe(undefined);
});
TestEnvironment.test("doesn't start HTML experiments or autocompletion when outcomes not preloaded", { clientContext: defaultClientContext }, function (env) {
    checkInit(env);
    jest.advanceTimersByTime(200);
    expect(env.htmlExperimentsStarted).toBe(false);
    expect(env.autocompleteCalled).toBe(false);
});
TestEnvironment.test("can run experiment when network fails", { clientContext: defaultClientContext }, function (env) {
    checkInit(env);
    var options = makeOptions(2);
    var ex = env.getClient().experiment(experimentName, options);
    expect(ex.pick).toMatch(/^o[01]$/);
    jest.advanceTimersByTime(200);
    if (env.allowSetLocalStorage) {
        checkState(env.localStorage, ex.pick);
    }
    env.checkStartExperiments(options, ex.pick, undefined);
});
test("html experiments and autocompletions begin when outcomes preloaded", function () {
    var preloadedOutcomes = {};
    var env = new TestEnvironment({}, defaultClientContext, true, true, true);
    var client = new Client_1.Client(env, appKey, function () { return null; }, preloadedOutcomes);
    env.setClient(client);
    expect(env.htmlExperimentsStarted).toBe(true);
    expect(env.autocompleteCalled).toBe(true);
});
TestEnvironment.test("single node", { outcomes: makeOutcomes(2, leaf(1)), clientContext: defaultClientContext }, function (env) {
    checkInit(env);
    var ex = env.getClient().experiment(experimentName, env.getOptions());
    expect(ex.pick).toBe("o1");
    jest.advanceTimersByTime(200);
    if (env.allowSetLocalStorage) {
        checkState(env.localStorage, "o1");
    }
    env.checkStartExperiments(env.getOptions(), "o1", true);
});
function testTree(name, outcomes, pick, clientContext) {
    if (clientContext === void 0) { clientContext = defaultClientContext; }
    var _loop_1 = function (complete) {
        TestEnvironment.test(name + (complete ? " with completion" : ""), { outcomes: outcomes, clientContext: clientContext }, function (env) {
            checkInit(env);
            var ex = env.getClient().experiment(experimentName, env.getOptions());
            expect(ex.pick).toBe(pick);
            jest.advanceTimersByTime(200);
            if (env.allowSetLocalStorage) {
                checkState(env.localStorage, pick);
            }
            var instanceKey = env.checkStartExperiments(env.getOptions(), pick, true);
            if (complete) {
                var callback = jest.fn();
                env.getClient().completeDefaults(0.123, callback);
                jest.advanceTimersByTime(150);
                expect(callback).toHaveBeenCalledTimes(1);
                checkCompleteExperiments(env.completeExperimentsData, instanceKey, pick, 0.123);
            }
        });
    };
    for (var _i = 0, _a = [false, true]; _i < _a.length; _i++) {
        var complete = _a[_i];
        _loop_1(complete);
    }
}
testTree("branch on tzo left", makeOutcomes(3, tzoLt(-410, leaf(1), leaf(2))), "o1");
testTree("branch on tzo right", makeOutcomes(3, tzoLt(-430, leaf(1), leaf(2))), "o2");
testTree("branch on lang left", makeOutcomes(3, langEq(defaultClientContext.lang, leaf(1), leaf(2))), "o1");
testTree("branch on lang right", makeOutcomes(3, langEq("not-" + defaultClientContext.lang, leaf(1), leaf(2))), "o2");
testTree("branch on undefined lang left", makeOutcomes(3, langEq("null", leaf(1), leaf(2))), "o1", { tzo: 0 });
testTree("branch on undefined lang right", makeOutcomes(3, langEq("not-null", leaf(1), leaf(2))), "o2", { tzo: 0 });
//# sourceMappingURL=Client.test.js.map